// Create an overlay mask with the motion vectors
// Copyright(c)2006 A.G.Balakhnin aka Fizick
// See legal notice in Copying.txt for more information

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#include "MaskFun.h"
#include <math.h>
#include <memory.h>
#include "CopyCode.h"

inline void ByteOccMask(BYTE *occMask, int occlusion, double occnorm, double fGamma)
{
	if (fGamma == 1.0)
		*occMask = max(*occMask, min((int)(255 * occlusion*occnorm),255));
	else
		*occMask= max(*occMask, min((int)(255 * pow(occlusion*occnorm, fGamma) ), 255));
}

void MakeVectorOcclusionMaskTime(MVClip &mvClip, int nBlkX, int nBlkY, double dMaskNormDivider, double fGamma, int nPel, BYTE * occMask, int occMaskPitch, int time256, int nBlkStepX, int nBlkStepY)
{ // analyse vectors field to detect occlusion
	MemZoneSet(occMask, 0, nBlkX, nBlkY, 0, 0, occMaskPitch);
	int time4096X = time256*16/(nBlkStepX*nPel); // blkstep here is really blksize-overlap
	int time4096Y = time256*16/(nBlkStepY*nPel);
	_asm emms;
	double occnormX = 80.0 / (dMaskNormDivider*nBlkStepX*nPel);
	double occnormY = 80.0 / (dMaskNormDivider*nBlkStepY*nPel);
    bool isBackward = mvClip.IsBackward();
	  int occlusion;

			  for (int by=0; by<nBlkY; by++)
			  {
				  for (int bx=0; bx<nBlkX; bx++)
				  {
					  int i = bx + by*nBlkX; // current block
						const FakeBlockData &block = mvClip.GetBlock(0, i);
						int vx = block.GetMV().x;
						int vy = block.GetMV().y;
						if (bx < nBlkX-1) // right neighbor
						{
							int i1 = i+1;
							const FakeBlockData &block1 = mvClip.GetBlock(0, i1);
							int vx1 = block1.GetMV().x;
							//int vy1 = block1.GetMV().y;
							if (vx1<vx) {
							    occlusion = vx-vx1;
							    int minb = isBackward ? max(0, bx+1-occlusion*time4096X/4096): bx;
							    int maxb = isBackward ? bx+1 : min(bx+1-occlusion*time4096X/4096, nBlkX-1);
                                for (int bxi=minb; bxi<=maxb; bxi++)
                                    ByteOccMask(&occMask[bxi+by*occMaskPitch], occlusion, occnormX, fGamma);
							}
						}
						if (by < nBlkY-1) // bottom neighbor
						{
							int i1 = i + nBlkX;
							const FakeBlockData &block1 = mvClip.GetBlock(0, i1);
							//int vx1 = block1.GetMV().x;
							int vy1 = block1.GetMV().y;
							if (vy1<vy) {
							    occlusion = vy-vy1;
							    int minb = isBackward ? max(0, by+1-occlusion*time4096Y/4096): by;
							    int maxb = isBackward ? by+1 : min(by+1-occlusion*time4096Y/4096, nBlkY-1);
                                for (int byi=minb; byi<=maxb; byi++)
                                     ByteOccMask(&occMask[bx+byi*occMaskPitch], occlusion, occnormY, fGamma);
							}
						}
				  }
			  }
}
unsigned char ByteNorm(unsigned int sad, double dSADNormFactor, double fGamma)
{
//	    double dSADNormFactor = 4 / (dMaskNormFactor*blkSizeX*blkSizeY);
	double l = 255 * pow(sad*dSADNormFactor, fGamma); // Fizick - now linear for gm=1
	return (unsigned char)((l > 255) ? 255 : l);
}

void MakeSADMaskTime(MVClip &mvClip, int nBlkX, int nBlkY, double dSADNormFactor, double fGamma, int nPel, BYTE * Mask, int MaskPitch, int time256, int nBlkStepX, int nBlkStepY)
{
// Make approximate SAD mask at intermediate time
//    double dSADNormFactor = 4 / (dMaskNormDivider*nBlkSizeX*nBlkSizeY);
    MemZoneSet(Mask, 0, nBlkX, nBlkY, 0, 0, MaskPitch);
    int time4096X = (256-time256)*16/(nBlkStepX*nPel); // blkstep here is really blksize-overlap
    int time4096Y = (256-time256)*16/(nBlkStepY*nPel);

    for (int by=0; by<nBlkY; by++)
    {
        for (int bx=0; bx<nBlkX; bx++)
        {
            int i = bx + by*nBlkX; // current block
            const FakeBlockData &block = mvClip.GetBlock(0, i);
            int vx = block.GetMV().x;
            int vy = block.GetMV().y;
            int bxi = bx - vx*time4096X/4096; // limits?
            int byi = by - vy*time4096Y/4096;
            if(bxi <0 || bxi >= nBlkX || byi <0 || byi >= nBlkY)
            {
                bxi = bx;
                byi = by;
            }
            int i1 = bxi + byi*nBlkX;
            int sad = mvClip.GetBlock(0, i1).GetSAD();
            Mask[bx + by*nBlkX] = ByteNorm(sad, dSADNormFactor, fGamma);
        }
    }
}
void MakeVectorSmallMasks(MVClip &mvClip, int nBlkX, int nBlkY, short *VXSmallY, int pitchVXSmallY, short *VYSmallY, int pitchVYSmallY)
{
	  // make  vector vx and vy small masks
	  for (int by=0; by<nBlkY; by++)
	  {
		  for (int bx=0; bx<nBlkX; bx++)
		  {
			  int i = bx + by*nBlkX;
				const FakeBlockData &block = mvClip.GetBlock(0, i);
				int vx = block.GetMV().x;
				int vy = block.GetMV().y;
				VXSmallY[bx+by*pitchVXSmallY] = vx; // luma
				VYSmallY[bx+by*pitchVYSmallY] = vy; // luma
		  }
	  }
}
void VectorSmallMaskYToHalfUV(short * VSmallY, int nBlkX, int nBlkY, short *VSmallUV, int ratioUV)
{
	if (ratioUV==2)
	{
	// YV12 colorformat
	  for (int by=0; by<nBlkY; by++)
	  {
		  for (int bx=0; bx<nBlkX; bx++)
		  {
				VSmallUV[bx] = ((VSmallY[bx])>>1); // chroma
		  }
		  VSmallY += nBlkX;
		  VSmallUV += nBlkX;
	  }
	}
	else // ratioUV==1
	{
	// Height YUY2 colorformat
	  for (int by=0; by<nBlkY; by++)
	  {
		  for (int bx=0; bx<nBlkX; bx++)
		  {
				VSmallUV[bx] = VSmallY[bx]; // chroma
		  }
		  VSmallY += nBlkX;
		  VSmallUV += nBlkX;
	  }
	}

}


void Merge4PlanesToBig(BYTE *pel2Plane, int pel2Pitch, const BYTE *pPlane0, const BYTE *pPlane1,
					  const BYTE *pPlane2, const BYTE * pPlane3, int width, int height, int pitch, bool isse)
{
	// copy refined planes to big one plane
	if (!isse)
	{
		for (int h=0; h<height; h++)
		{
			for (int w=0; w<width; w++)
			{
				pel2Plane[w<<1] = pPlane0[w];
				pel2Plane[(w<<1) +1] = pPlane1[w];
			}
			pel2Plane += pel2Pitch;
			for (int w=0; w<width; w++)
			{
				pel2Plane[w<<1] = pPlane2[w];
				pel2Plane[(w<<1) +1] = pPlane3[w];
			}
			pel2Plane += pel2Pitch;
			pPlane0 += pitch;
			pPlane1 += pitch;
			pPlane2 += pitch;
			pPlane3 += pitch;
		}
	}
	else // isse - not very optimized
	{
		_asm
		{
			push ebx;
			push esi;
			push edi;
			mov edi, pel2Plane;
			mov esi, pPlane0;
			mov edx, pPlane1;

			mov ebx, height;
looph1:
			mov ecx, width;
			mov eax, 0;
align 16
loopw1:
			movd mm0, [esi+eax];
			movd mm1, [edx+eax];
			punpcklbw mm0, mm1;
			shl eax, 1;
			movq [edi+eax], mm0;
			shr eax, 1;
			add eax, 4;
			cmp eax, ecx;
			jl loopw1;

			mov eax, pel2Pitch;
			add edi, eax;
			add edi, eax;
			mov eax, pitch;
			add esi, eax;
			add edx, eax;
			dec ebx;
			cmp ebx, 0;
			jg looph1;

			mov edi, pel2Plane;
			mov esi, pPlane2;
			mov edx, pPlane3;

			mov eax, pel2Pitch;
			add edi, eax;

			mov ebx, height;
looph2:
			mov ecx, width;
			mov eax, 0;
align 16
loopw2:
			movd mm0, [esi+eax];
			movd mm1, [edx+eax];
			punpcklbw mm0, mm1;
			shl eax, 1;
			movq [edi+eax], mm0;
			shr eax, 1;
			add eax, 4;
			cmp eax, ecx;
			jl loopw2;

			mov eax, pel2Pitch;
			add edi, eax;
			add edi, eax;
			mov eax, pitch;
			add esi, eax;
			add edx, eax;
			dec ebx;
			cmp ebx, 0;
			jg looph2;

			pop edi;
			pop esi;
			pop ebx;
			emms;

		}
	}
}


void Merge16PlanesToBig(BYTE *pel4Plane, int pel4Pitch,
                    const BYTE *pPlane0, const BYTE *pPlane1, const BYTE *pPlane2, const BYTE * pPlane3,
                    const BYTE *pPlane4, const BYTE *pPlane5, const BYTE *pPlane6, const BYTE * pPlane7,
                    const BYTE *pPlane8, const BYTE *pPlane9, const BYTE *pPlane10, const BYTE * pPlane11,
                    const BYTE *pPlane12, const BYTE * pPlane13, const BYTE *pPlane14, const BYTE * pPlane15,
					int width, int height, int pitch, bool isse)
{
	// copy refined planes to big one plane
//	if (!isse)
	{
		for (int h=0; h<height; h++)
		{
			for (int w=0; w<width; w++)
			{
				pel4Plane[w<<2] = pPlane0[w];
				pel4Plane[(w<<2) +1] = pPlane1[w];
				pel4Plane[(w<<2) +2] = pPlane2[w];
				pel4Plane[(w<<2) +3] = pPlane3[w];
			}
			pel4Plane += pel4Pitch;
			for (int w=0; w<width; w++)
			{
				pel4Plane[w<<2] = pPlane4[w];
				pel4Plane[(w<<2) +1] = pPlane5[w];
				pel4Plane[(w<<2) +2] = pPlane6[w];
				pel4Plane[(w<<2) +3] = pPlane7[w];
			}
			pel4Plane += pel4Pitch;
			for (int w=0; w<width; w++)
			{
				pel4Plane[w<<2] = pPlane8[w];
				pel4Plane[(w<<2) +1] = pPlane9[w];
				pel4Plane[(w<<2) +2] = pPlane10[w];
				pel4Plane[(w<<2) +3] = pPlane11[w];
			}
			pel4Plane += pel4Pitch;
			for (int w=0; w<width; w++)
			{
				pel4Plane[w<<2] = pPlane12[w];
				pel4Plane[(w<<2) +1] = pPlane13[w];
				pel4Plane[(w<<2) +2] = pPlane14[w];
				pel4Plane[(w<<2) +3] = pPlane15[w];
			}
			pel4Plane += pel4Pitch;
			pPlane0 += pitch;
			pPlane1 += pitch;
			pPlane2 += pitch;
			pPlane3 += pitch;
			pPlane4 += pitch;
			pPlane5 += pitch;
			pPlane6 += pitch;
			pPlane7 += pitch;
			pPlane8 += pitch;
			pPlane9 += pitch;
			pPlane10 += pitch;
			pPlane11 += pitch;
			pPlane12 += pitch;
			pPlane13+= pitch;
			pPlane14 += pitch;
			pPlane15 += pitch;
		}
	}
}

//-----------------------------------------------------------
unsigned char SADToMask(unsigned int sad, unsigned int sadnorm1024)
{
	// sadnorm1024 = 255 * (4*1024)/(mlSAD*nBlkSize*nBlkSize*chromablockfactor)
	unsigned int l = sadnorm1024*sad/1024;
	return (unsigned char)((l > 255) ? 255 : l);
}


// time-weihted blend src with ref frames (used for interpolation for poor motion estimation)
void Blend(BYTE * pdst, const BYTE * psrc, const BYTE * pref, int height, int width, int dst_pitch, int src_pitch, int ref_pitch, int time256, bool isse)
{
    // add isse
    int h, w;
        for (h=0; h<height; h++)
        {
            for (w=0; w<width; w++)
            {
                pdst[w] = (psrc[w]*(256 - time256) + pref[w]*time256)>>8;
            }
            pdst += dst_pitch;
            psrc += src_pitch;
            pref += ref_pitch;
        }
}
void FlowInter(BYTE * pdst, int dst_pitch, const BYTE *prefB, const BYTE *prefF, int ref_pitch,
			   short *VXFullB, short *VXFullF, short *VYFullB, short *VYFullF, BYTE *MaskB, BYTE *MaskF,
			   int VPitch, int width, int height, int time256, int nPel)
{
	if (nPel==1)
	{
		for (int h=0; h<height; h++)
		{
			for (int w=0; w<width; w++)
			{
				int vxF = (VXFullF[w]*time256)>>8;
				int vyF = (VYFullF[w]*time256)>>8;
				int dstF = prefF[vyF*ref_pitch + vxF + w];
				int dstF0 = prefF[w]; // zero
				int vxB = (VXFullB[w]*(256-time256))>>8;
				int vyB = (VYFullB[w]*(256-time256))>>8;
				int dstB = prefB[vyB*ref_pitch + vxB + w];
				int dstB0 = prefB[w]; // zero
				pdst[w] = ( ( (dstF*(255-MaskF[w]) + ((MaskF[w]*(dstB*(255-MaskB[w])+MaskB[w]*dstF0)+255)>>8) + 255)>>8 )*(256-time256) +
				            ( (dstB*(255-MaskB[w]) + ((MaskB[w]*(dstF*(255-MaskF[w])+MaskF[w]*dstB0)+255)>>8) + 255)>>8 )*time256 )>>8;
			}
			pdst += dst_pitch;
			prefB += ref_pitch;
			prefF += ref_pitch;
			VXFullB += VPitch;
			VYFullB += VPitch;
			VXFullF += VPitch;
			VYFullF += VPitch;
			MaskB += VPitch;
			MaskF += VPitch;
		}
	}
	else if (nPel==2)
	{
		for (int h=0; h<height; h++)
		{
			for (int w=0; w<width; w++)
			{
				int vxF = (VXFullF[w]*time256)>>8;
				int vyF = (VYFullF[w]*time256)>>8;
				int dstF = prefF[vyF*ref_pitch + vxF + (w<<1)];
				int dstF0 = prefF[(w<<1)]; // zero
				int vxB = (VXFullB[w]*(256-time256))>>8;
				int vyB = (VYFullB[w]*(256-time256))>>8;
				int dstB = prefB[vyB*ref_pitch + vxB + (w<<1)];
				int dstB0 = prefB[(w<<1)]; // zero
				pdst[w] = ( ( (dstF*(255-MaskF[w]) + ((MaskF[w]*(dstB*(255-MaskB[w])+MaskB[w]*dstF0)+255)>>8) + 255)>>8 )*(256-time256) +
				            ( (dstB*(255-MaskB[w]) + ((MaskB[w]*(dstF*(255-MaskF[w])+MaskF[w]*dstB0)+255)>>8) + 255)>>8 )*time256 )>>8;
//				pdst[w] = ( ( (dstF*(255-MaskF[w]) + dstB*MaskF[w] + 255)>>8 )*(256-time256) +
//				            ( (dstB*(255-MaskB[w]) + dstF*MaskB[w] + 255)>>8 )*time256 )>>8;
			}
			pdst += dst_pitch;
			prefB += ref_pitch<<1;
			prefF += ref_pitch<<1;
			VXFullB += VPitch;
			VYFullB += VPitch;
			VXFullF += VPitch;
			VYFullF += VPitch;
			MaskB += VPitch;
			MaskF += VPitch;
		}
	}
	else if (nPel==4)
	{
		for (int h=0; h<height; h++)
		{
			for (int w=0; w<width; w++)
			{
				int vxF = (VXFullF[w]*time256)>>8;
				int vyF = (VYFullF[w]*time256)>>8;
				int dstF = prefF[vyF*ref_pitch + vxF + (w<<2)];
				int dstF0 = prefF[(w<<2)]; // zero
				int vxB = (VXFullB[w]*(256-time256))>>8;
				int vyB = (VYFullB[w]*(256-time256))>>8;
				int dstB = prefB[vyB*ref_pitch + vxB + (w<<2)];
				int dstB0 = prefB[(w<<2)]; // zero
				pdst[w] = ( ( (dstF*(255-MaskF[w]) + ((MaskF[w]*(dstB*(255-MaskB[w])+MaskB[w]*dstF0)+255)>>8) + 255)>>8 )*(256-time256) +
				            ( (dstB*(255-MaskB[w]) + ((MaskB[w]*(dstF*(255-MaskF[w])+MaskF[w]*dstB0)+255)>>8) + 255)>>8 )*time256 )>>8;
			}
			pdst += dst_pitch;
			prefB += ref_pitch<<2;
			prefF += ref_pitch<<2;
			VXFullB += VPitch;
			VYFullB += VPitch;
			VXFullF += VPitch;
			VYFullF += VPitch;
			MaskB += VPitch;
			MaskF += VPitch;
		}
	}
}

// FIND MEDIAN OF 3 ELEMENTS
//
inline int Median3 (int a, int b, int c)
{
	// b a c || c a b
	if (((b <= a) && (a <= c)) || ((c <= a) && (a <= b))) return a;

	// a b c || c b a
	else if (((a <= b) && (b <= c)) || ((c <= b) && (b <= a))) return b;

	// b c a || a c b
	else return c;

}

inline int Median3r (int a, int b, int c)
{
    // reduced median - if it is known that a <= c (more fast)

	// b a c
	if (b <= a) return a;
	// a c b
	else  if (c <= b) return c;
	// a b c
	else return b;

}
void FlowInterExtra(BYTE * pdst, int dst_pitch, const BYTE *prefB, const BYTE *prefF, int ref_pitch,
			   short *VXFullB, short *VXFullF, short *VYFullB, short *VYFullF, BYTE *MaskB, BYTE *MaskF,
			   int VPitch, int width, int height, int time256, int nPel,
			   short *VXFullBB, short *VXFullFF, short *VYFullBB, short *VYFullFF)
{
 	if (nPel==1)
	{
		for (int h=0; h<height; h++)
		{
			for (int w=0; w<width; w++)
			{
				int vxF = (VXFullF[w]*time256)>>8;
				int vyF = (VYFullF[w]*time256)>>8;
				int adrF = vyF*ref_pitch + vxF + w;
				int dstF = prefF[adrF];

				int vxFF = (VXFullFF[w]*time256)>>8; // is it correct time?
				int vyFF = (VYFullFF[w]*time256)>>8;
				int adrFF = vyFF*ref_pitch + vxFF + w;
				int dstFF = prefF[adrFF];

				int vxB = (VXFullB[w]*(256-time256))>>8;
				int vyB = (VYFullB[w]*(256-time256))>>8;
				int adrB = vyB*ref_pitch + vxB + w;
				int dstB = prefB[adrB];

				int vxBB = (VXFullBB[w]*(256-time256))>>8;
				int vyBB = (VYFullBB[w]*(256-time256))>>8;
				int adrBB = vyBB*ref_pitch + vxBB + w;
				int dstBB = prefB[adrBB];

                // use median, firstly get min max of compensations
                int minfb;
                int maxfb;
                if (dstF > dstB) {
                    minfb = dstB;
                    maxfb = dstF;
                } else {
                    maxfb = dstB;
                    minfb = dstF;
                }

                pdst[w] = (((Median3r(minfb,dstBB,maxfb)*MaskF[w] + dstF*(255-MaskF[w])+255)>>8)*(256-time256) +
                           ((Median3r(minfb,dstFF,maxfb)*MaskB[w] + dstB*(255-MaskB[w])+255)>>8)*time256)>>8;
  			}
			pdst += dst_pitch;
			prefB += ref_pitch;
			prefF += ref_pitch;
			VXFullB += VPitch;
			VYFullB += VPitch;
			VXFullF += VPitch;
			VYFullF += VPitch;
			MaskB += VPitch;
			MaskF += VPitch;
			VXFullBB += VPitch;
			VYFullBB += VPitch;
			VXFullFF += VPitch;
			VYFullFF += VPitch;
		}
	}
	else if (nPel==2)
	{
		for (int h=0; h<height; h++)
		{
			for (int w=0; w<width; w++)
			{
				int vxF = (VXFullF[w]*time256)>>8;
				int vyF = (VYFullF[w]*time256)>>8;
				int adrF = vyF*ref_pitch + vxF + (w<<1);
				int dstF = prefF[adrF];

				int vxFF = (VXFullFF[w]*time256)>>8; // is it correct time?
				int vyFF = (VYFullFF[w]*time256)>>8;
				int adrFF = vyFF*ref_pitch + vxFF + (w<<1);
				int dstFF = prefF[adrFF];

				int vxB = (VXFullB[w]*(256-time256))>>8;
				int vyB = (VYFullB[w]*(256-time256))>>8;
				int adrB = vyB*ref_pitch + vxB + (w<<1);
				int dstB = prefB[adrB];

				int vxBB = (VXFullBB[w]*(256-time256))>>8;
				int vyBB = (VYFullBB[w]*(256-time256))>>8;
				int adrBB = vyBB*ref_pitch + vxBB + (w<<1);
				int dstBB = prefB[adrBB];
                // use median, firsly get min max of compensations
                int minfb;
                int maxfb;
                if (dstF > dstB) {
                    minfb = dstB;
                    maxfb = dstF;
                } else {
                    maxfb = dstB;
                    minfb = dstF;
                }

                pdst[w] = (((Median3r(minfb, dstBB, maxfb)*MaskF[w] + dstF*(255-MaskF[w])+255)>>8)*(256-time256) +
                           ((Median3r(minfb, dstFF, maxfb)*MaskB[w] + dstB*(255-MaskB[w])+255)>>8)*time256)>>8;

			}
			pdst += dst_pitch;
			prefB += ref_pitch<<1;
			prefF += ref_pitch<<1;
			VXFullB += VPitch;
			VYFullB += VPitch;
			VXFullF += VPitch;
			VYFullF += VPitch;
			MaskB += VPitch;
			MaskF += VPitch;
			VXFullBB += VPitch;
			VYFullBB += VPitch;
			VXFullFF += VPitch;
			VYFullFF += VPitch;
		}
	}
	else if (nPel==4)
	{
		for (int h=0; h<height; h++)
		{
			for (int w=0; w<width; w++)
			{
				int vxF = (VXFullF[w]*time256)>>8;
				int vyF = (VYFullF[w]*time256)>>8;
				int adrF = vyF*ref_pitch + vxF + (w<<2);
				int dstF = prefF[adrF];

				int vxFF = (VXFullFF[w]*time256)>>8; // is it correct time?
				int vyFF = (VYFullFF[w]*time256)>>8;
				int adrFF = vyFF*ref_pitch + vxFF + (w<<2);
				int dstFF = prefF[adrFF];

				int vxB = (VXFullB[w]*(256-time256))>>8;
				int vyB = (VYFullB[w]*(256-time256))>>8;
				int adrB = vyB*ref_pitch + vxB + (w<<2);
				int dstB = prefB[adrB];

				int vxBB = (VXFullBB[w]*(256-time256))>>8;
				int vyBB = (VYFullBB[w]*(256-time256))>>8;
				int adrBB = vyBB*ref_pitch + vxBB + (w<<2);
				int dstBB = prefB[adrBB];
                // use median, firsly get min max of compensations
                int minfb;
                int maxfb;
                if (dstF > dstB) {
                    minfb = dstB;
                    maxfb = dstF;
                } else {
                    maxfb = dstB;
                    minfb = dstF;
                }

                pdst[w] = (((Median3r(minfb, dstBB, maxfb)*MaskF[w] + dstF*(255-MaskF[w])+255)>>8)*(256-time256) +
                           ((Median3r(minfb, dstFF, maxfb)*MaskB[w] + dstB*(255-MaskB[w])+255)>>8)*time256)>>8;
			}
			pdst += dst_pitch;
			prefB += ref_pitch<<2;
			prefF += ref_pitch<<2;
			VXFullB += VPitch;
			VYFullB += VPitch;
			VXFullF += VPitch;
			VYFullF += VPitch;
			MaskB += VPitch;
			MaskF += VPitch;
			VXFullBB += VPitch;
			VYFullBB += VPitch;
			VXFullFF += VPitch;
			VYFullFF += VPitch;
		}
	}
}
void FlowInterSimple(BYTE * pdst, int dst_pitch, const BYTE *prefB, const BYTE *prefF, int ref_pitch,
			   short *VXFullB, short *VXFullF, short *VYFullB, short *VYFullF, BYTE *MaskB, BYTE *MaskF,
			   int VPitch, int width, int height, int time256, int nPel)
{

    if (time256==128) // special case double fps - fastest
    {
        if (nPel==1)
        {
            for (int h=0; h<height; h++)
            {
                for (int w=0; w<width; w+=2) // paired for speed
                {
    				int vxF = (VXFullF[w])>>1;
    				int vyF = (VYFullF[w])>>1;
                    int addrF = vyF*ref_pitch + vxF + w;
                    int dstF = prefF[addrF];
                    int dstF1 = prefF[addrF+1]; // approximation for speed
    				int vxB = (VXFullB[w])>>1;
    				int vyB = (VYFullB[w])>>1;
                    int addrB = vyB*ref_pitch + vxB + w;
                    int dstB = prefB[addrB];
                    int dstB1 = prefB[addrB+1];
                    pdst[w] = ( ((dstF + dstB)<<8) + (dstB - dstF)*(MaskF[w] - MaskB[w]) )>>9;
                    pdst[w+1] = ( ((dstF1 + dstB1)<<8) + (dstB1 - dstF1)*(MaskF[w+1] - MaskB[w+1]) )>>9;
                }
                pdst += dst_pitch;
                prefB += ref_pitch;
                prefF += ref_pitch;
                VXFullB += VPitch;
                VYFullB += VPitch;
                VXFullF += VPitch;
                VYFullF += VPitch;
                MaskB += VPitch;
                MaskF += VPitch;
            }
        }
        else if (nPel==2)
        {
            for (int h=0; h<height; h++)
            {
                for (int w=0; w<width; w+=1)
                {
    				int vxF = (VXFullF[w])>>1;
    				int vyF = (VYFullF[w])>>1;
                    int dstF = prefF[vyF*ref_pitch + vxF + (w<<1)];
    				int vxB = (VXFullB[w])>>1;
    				int vyB = (VYFullB[w])>>1;
                    int dstB = prefB[vyB*ref_pitch + vxB + (w<<1)];
                    pdst[w] = ( ((dstF + dstB)<<8) + (dstB - dstF)*(MaskF[w] - MaskB[w]) )>>9;
                }
                pdst += dst_pitch;
                prefB += ref_pitch<<1;
                prefF += ref_pitch<<1;
                VXFullB += VPitch;
                VYFullB += VPitch;
                VXFullF += VPitch;
                VYFullF += VPitch;
                MaskB += VPitch;
                MaskF += VPitch;
            }
        }
        else if (nPel==4)
        {
            for (int h=0; h<height; h++)
            {
                for (int w=0; w<width; w+=1)
                {
    				int vxF = (VXFullF[w])>>1;
    				int vyF = (VYFullF[w])>>1;
                    int dstF = prefF[vyF*ref_pitch + vxF + (w<<2)];
    				int vxB = (VXFullB[w])>>1;
    				int vyB = (VYFullB[w])>>1;
                    int dstB = prefB[vyB*ref_pitch + vxB + (w<<2)];
                    pdst[w] = ( ((dstF + dstB)<<8) + (dstB - dstF)*(MaskF[w] - MaskB[w]) )>>9;
                }
                pdst += dst_pitch;
                prefB += ref_pitch<<2;
                prefF += ref_pitch<<2;
                VXFullB += VPitch;
                VYFullB += VPitch;
                VXFullF += VPitch;
                VYFullF += VPitch;
                MaskB += VPitch;
                MaskF += VPitch;
            }
        }
    }
    else // general case
    {
        if (nPel==1)
        {
            for (int h=0; h<height; h++)
            {
                for (int w=0; w<width; w+=2) // paired for speed
                {
                    int vxF = (VXFullF[w]*time256)>>8;
                    int vyF = (VYFullF[w]*time256)>>8;
                    int addrF = vyF*ref_pitch + vxF + w;
                    int dstF = prefF[addrF];
                    int dstF1 = prefF[addrF+1]; // approximation for speed
                    int vxB = (VXFullB[w]*(256-time256))>>8;
                    int vyB = (VYFullB[w]*(256-time256))>>8;
                    int addrB = vyB*ref_pitch + vxB + w;
                    int dstB = prefB[addrB];
                    int dstB1 = prefB[addrB+1];
                    pdst[w] = ( ( (dstF*255 + (dstB-dstF)*MaskF[w] + 255) )*(256-time256) +
                                ( (dstB*255 - (dstB-dstF)*MaskB[w] + 255) )*time256 )>>16;
                    pdst[w+1] = ( ( (dstF1*255 + (dstB1-dstF1)*MaskF[w+1] + 255) )*(256-time256) +
                                ( (dstB1*255 - (dstB1-dstF1)*MaskB[w+1] + 255) )*time256 )>>16;
                }
                pdst += dst_pitch;
                prefB += ref_pitch;
                prefF += ref_pitch;
                VXFullB += VPitch;
                VYFullB += VPitch;
                VXFullF += VPitch;
                VYFullF += VPitch;
                MaskB += VPitch;
                MaskF += VPitch;
            }
        }
        else if (nPel==2)
        {
            for (int h=0; h<height; h++)
            {
                for (int w=0; w<width; w+=1)
                {
    				int vxF = (VXFullF[w]*time256)>>8;
    				int vyF = (VYFullF[w]*time256)>>8;
                    int dstF = prefF[vyF*ref_pitch + vxF + (w<<1)];
                    int vxB = (VXFullB[w]*(256-time256))>>8;
                    int vyB = (VYFullB[w]*(256-time256))>>8;
                    int dstB = prefB[vyB*ref_pitch + vxB + (w<<1)];
                    pdst[w] = ( ( (dstF*(255-MaskF[w]) + dstB*MaskF[w] + 255)>>8 )*(256-time256) +
                                ( (dstB*(255-MaskB[w]) + dstF*MaskB[w] + 255)>>8 )*time256 )>>8;
                }
                pdst += dst_pitch;
                prefB += ref_pitch<<1;
                prefF += ref_pitch<<1;
                VXFullB += VPitch;
                VYFullB += VPitch;
                VXFullF += VPitch;
                VYFullF += VPitch;
                MaskB += VPitch;
                MaskF += VPitch;
            }
        }
        else if (nPel==4)
        {
            for (int h=0; h<height; h++)
            {
                for (int w=0; w<width; w+=1)
                {
    				int vxF = (VXFullF[w]*time256)>>8;
    				int vyF = (VYFullF[w]*time256)>>8;
                    int dstF = prefF[vyF*ref_pitch + vxF + (w<<2)];
                    int vxB = (VXFullB[w]*(256-time256))>>8;
                    int vyB = (VYFullB[w]*(256-time256))>>8;
                    int dstB = prefB[vyB*ref_pitch + vxB + (w<<2)];
                    pdst[w] = ( ( (dstF*(255-MaskF[w]) + dstB*MaskF[w] + 255)>>8 )*(256-time256) +
                                ( (dstB*(255-MaskB[w]) + dstF*MaskB[w] + 255)>>8 )*time256 )>>8;
                }
                pdst += dst_pitch;
                prefB += ref_pitch<<2;
                prefF += ref_pitch<<2;
                VXFullB += VPitch;
                VYFullB += VPitch;
                VXFullF += VPitch;
                VYFullF += VPitch;
                MaskB += VPitch;
                MaskF += VPitch;
            }
        }
    }
}
